"""ExtFile product module."""
#CHANGES FOR CMF COMPATIBILITY: SEARCH FOR "CMF-CHANGE"
#jms/JanixCorp/jeff@janix.com/3.16.2001
#note: tabs have been changed to spaces in this file.
###############################################################################
#
# Copyright (c) 2000 Gregor Heine and Communications & Design Consultancy GmbH,
# Solmsstrasse 2, 60486 Frankfurt, Germany (http://www.cdc-group.com)
# All rights reserved. Written by Gregor Heine <heine@cdc-group.com>
# ExtFile Home: http://www.zope.org/Members/MacGregor/ExtFile/index_html
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#   
#  In accordance with the license provided for by the software upon
#  which some of the source code has been derived or used, the following
#  acknowledgement is hereby provided :
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
###############################################################################

__doc__ = """
    ExtFile product module.
    The ExtFile-Product works like the Zope File-product, but stores the 
    uploaded file externally in a repository-direcory.

$Id: ExtFile.py,v 2.03 2001/02/26 14:43:58 heine Exp $"""

__version__='$Release: 1.1.0.beta3 $'[10:-2]

import Globals
from Globals import HTMLFile, MessageDialog, Persistent
from OFS.PropertyManager import PropertyManager
from OFS.FindSupport import FindSupport
from OFS.content_types import guess_content_type
from webdav.common import rfc1123_date
import OFS.SimpleItem
import Acquisition
import AccessControl.Role
from __main__ import *
import Zope
from Products.ZCatalog.CatalogAwareness import CatalogAware
from DateTime import DateTime
from types import StringType
import urllib, os, string
from os.path import join, isfile
try: from cStringIO import StringIO
except: from StringIO import StringIO




NO_BACKUPS = 0
BACKUP_ON_DELETE = 1
ALWAYS_BACKUP = 2
UNDO_POLICY = BACKUP_ON_DELETE
bad =' ()[]{}ݟ'
good='_______AAAAAAaaaaaaCcEEEEeeeeeIIIIiiiiNnOOOOOOooooooSssUUUUuuuuYYyyZz'
TRANSMAP=string.maketrans(bad, good)

manage_addExtFileForm = HTMLFile('extFileAdd', globals()) 

def manage_addExtFile(self, id='', title='', descr='', file='', content_type='', permission_check=0, REQUEST=None):
    """Add a ExtFile to a folder."""
#   print self._d.getPhysicalPath()
    if id=='' and type(file) is not type('') and hasattr(file,'filename'): 
        # generate id from filename and make sure, there are no spaces in the id
        id=file.filename
        id=id[max(string.rfind(id,'/'), 
                  string.rfind(id,'\\'), 
                  string.rfind(id,':')
                 )+1:]
        id=string.translate(id, TRANSMAP)
    if id:
        self=self.this()
        tempExtFile=ExtFile(id, title, descr, permission_check)
        self._setObject(id, tempExtFile)
        self._getOb(id).manage_file_upload(file, content_type)
        if REQUEST is not None:
            return MessageDialog(title = 'Created',
                message = 'The ExtFile %s was successfully created!' % id,
                action = './manage_main',)
    elif REQUEST is not None:
        return MessageDialog(title = 'Error!',
            message = 'You must specify a file!', action = './manage_main',)

class ExtFile(
        CatalogAware,                   # ZCatalog Awareness
        OFS.SimpleItem.Item,            # A simple Principia object. Not Folderish. 
        Persistent,                     # Make us persistent. Yaah! 
        Acquisition.Implicit,           # Let us use our parent's variables and methods. 
        AccessControl.Role.RoleManager, # Security manager. 
        PropertyManager
    ): 
    
    _properties=(                       # what properties have we?
        {'id':'title',                          'type':'string'},
        {'id':'descr',                          'type':'string'},
        {'id':'content_type',                   'type':'string'},
        {'id':'use_download_permission_check',  'type':'int'},
    )

    manage_options = (                  # what management options are there?
        {'label':'Edit',            'action': 'manage_main'             },
#       {'label':'Meta Data',       'action': 'view'                    },
        {'label':'View/Download',   'action': ''                        },
        {'label':'Upload',          'action': 'manage_uploadForm'       },
        {'label':'Properties',      'action': 'manage_propertiesForm'   },
        {'label':'Security',        'action': 'manage_access'           },
    ) 

    __ac_permissions__=(                # what permissions make sense for us? 
        ('View management screens', ('manage_tabs',
                                     'manage_main',
                                     'manage_uploadForm')),
        ('Change permissions',      ('manage_access',)),
        ('Change ExtFile/ExtImage', ('manage_editExtFile',
                                     'manage_file_upload',
                                     'manage_http_upload',
                                     'PUT')),
        ('FTP access',              ('manage_FTPstat',
                                     'manage_FTPget',
                                     'manage_FTPlist')),
        ('Download ExtFile/ExtImage',()),
        ('View',                    ('index_html',
                                     'view',
                                     'icon_html',
                                     'icon_gif',
                                     'link',
                                     'is_broken',
                                     'get_size',
                                     'getContentType',
                                     '__str__'
                                     )),
    )

    default_catalog="catalog"
    
    meta_type = 'ExtFile_CMF'               # what do people think they're adding? 
    
    _repository =['var','reposit']      # location of the file-repository

    use_download_permission_check=0

    # MIME-Type Dictionary. To add a MIME-Type, add a file in the directory 
    # icons/_category_/_subcategory-icon-file_
    # example: Icon tifficon.gif for the MIME-Type image/tiff goes into 
    # icons/image/tifficon.gif and the dictionary must be updated like this: 
    # 'image':{'tiff':'tifficon.gif','default':'default.gif'}, ...
    _types={'image':                    
                {'default':'default.gif'},
            'text':
                {'html':'html.gif', 'xml':'xml.gif', 'default':'default.gif', 
                 'python':'py.gif'},
            'application':
                {'pdf':'pdf.gif', 'zip':'zip.gif', 'tar':'zip.gif', 
                 'msword':'doc.gif', 'excel':'xls.gif', 'powerpoint':'ppt.gif', 
                 'default':'default.gif'},
            'video':
                {'default':'default.gif'},
            'audio':
                {'default':'default.gif'},
            'default':'default.gif'
        }

    ################################
    # Init method                  #
    ################################

    def __init__(self, id, title='', descr='', permission_check=0): 
        """initialize a new instance of ExtFile"""
        reposit=self._reposit_path()
        self.id = id
        self.title = title
        self.descr=descr
        self.use_download_permission_check=permission_check
        self.__version__=__version__
        if os.path.exists(reposit)!=1:
            mkdirExc='Cannot create directory: '
            try:
                os.mkdir(reposit)
            except:
                raise mkdirExc, reposit

        #CMF-CHANGE add more dirs, one for each [a-z]
        z = join(reposit,'z')
        if os.path.exists(z)!=1:    
            mksubdirExc='Cannot create SUB-directory: '
            letters = string.lowercase[:26]
            for letter in letters:
                p = join(reposit,letter)
                if os.path.exists(p)!=1:
                    try:
                        os.mkdir(p)
                    except:
                        raise mksubdirExc, p
                    
                            
                
                
                
                
        self.filename=''
        self.content_type= ''
            
    ################################
    # Public methods               #
    ################################

    #CMF-CHANGE : had to remove this since it mixed up the portal interface
    #view = HTMLFile('viewExtFile', globals())          # Public HTML Interface
    
    def index_html (self, REQUEST, RESPONSE, icon=None, preview=0, width=None, height=None):
        """return the file with it's corresponding MIME-type"""
        # HTTP If-Modified-Since header handling. (Copied from OFS/Image.py)
        header=REQUEST.get_header('If-Modified-Since', None)
        if header is not None:
            header=string.split(header, ';')[0]
            try:    mod_since=long(DateTime(header).timeTime())
            except: mod_since=None
            if mod_since is not None:
                if self._p_mtime:
                    last_mod = long(self._p_mtime)
                else:
                    last_mod = long(0)
                if last_mod > 0 and last_mod < mod_since:
                    RESPONSE.setStatus(304)
                    return ''
        
        if hasattr(self,'has_preview') and self.has_preview: has_preview=1
        else: has_preview=0
        if not self._access_permitted(REQUEST): preview=1
        if (preview and not has_preview): icon=1
        
        if icon:
            prodpath=join(SOFTWARE_HOME,'Products','CMFOptions','ExtFile')
            content_type='image/gif'
            filename=join(prodpath,self.getIconPath())
        elif preview:
            filename=join(self._reposit_path(),self.prev_filename)
            content_type=self.prev_content_type
        else:
            filename=join(self._reposit_path(),self.filename)
            content_type=self.content_type
        
        cantReadExc = 'Cannot read (?did you upload?): '
        try:
            if not isfile(filename): self._undo()
            size=os.stat(filename)[6] # file size
        except:
            filename=join(SOFTWARE_HOME,'Products','CMFOptions','ExtFile','icons','broken.gif')
            try: size=os.stat(filename)[6]
            except: raise cantReadExc, ("%s (%s)" %(self.id, filename))
            content_type='image/gif'
        data=StringIO()
        if None and width is not None and height is not None:
            try:
                from PIL import Image
                im = Image.open(filename) 
                if im.mode!='RGB' and im.mode!='CMYK': im=im.convert("RGB")
                im.draft(None,(int(width),int(height)))
                im.load()
                im = im.resize((int(width),int(height)), Image.BICUBIC)
                im.save(data, 'JPEG')
            except:
                try:
                    self._copyfile(filename, data)
                except:
                    raise cantReadExc, ("%s (%s)" %(self.id, filename))
            else:
                data.seek(0,2)
                size=data.tell()
                content_type='image/jpeg'
        else:
            try:
                self._copyfile(filename, data)
            except:
                raise cantReadExc, ("%s (%s)" %(self.id, filename))
        RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
        RESPONSE.setHeader('Content-Type', content_type)
        RESPONSE.setHeader('Content-Length', size)
        return data.getvalue()
    
    def __len__(self): return 1

    def __str__(self): return self.link()

    def link(self, text='', **args):
        """return a HTML link tag to the file"""
        if text=='': text=self.title_or_id()
        strg='<a href="%s"' % (self.absolute_url())
        for key in args.keys():
            value = args.get(key)
            strg = '%s %s="%s"' % (strg, key, value)
        strg = '%s>%s</a>' % (strg, text)
        return strg
    
    def icon_gif(self, REQUEST, RESPONSE):
        """return an icon for the file's MIME-Type"""
        REQUEST.RESPONSE.redirect(REQUEST['URL1']+'?icon=1')
        return ''
    
    def icon_html(self):
        """The icon embedded in html with a link to the real file"""
        strg = '<IMG SRC="%s?icon=1" BORDER="0">' % (self.absolute_url())
        if self._access_permitted(self.REQUEST): return self.link(strg)
        else: return strg
    
    def is_broken(self):
        """Check if external file exists and return true (1) or false (0)"""
        fn=join(self._reposit_path(),self.filename)
        if not isfile(fn):
            self._undo()
            if not isfile(fn):
                return 1
        return 0
    
    def get_size(self):
        """Returns the size of the file or image."""
        fn=join(self._reposit_path(),self.filename)
        try:
            size=os.stat(fn)[6]
        except:
            try: size=os.stat(fn+'.undo')[6]
            except: size=0
        return size
    
    rawsize=get_size
    getSize=get_size
    
    def size(self):
        """Returns a formatted stringified version of the file size"""
        return self._bytetostring(self.rawsize())

    def getContentType(self):
        """Get the content type of a file or image.
        Returns the content type (MIME type) of a file or image."""
        return self.content_type
        
    def _getMIMECatAndSub(self, mime_string):
        """Split MIME String into Category and Subcategory"""
        cat=mime_string[:string.find(mime_string, '/')]     # MIME-category
        sub=mime_string[string.find(mime_string, '/')+1:]       # sub-category
        return cat, sub

    def getIconPath(self):
        """Depending on the MIME Type of the file/image an icon
           can be displayed. This function determines which
           image in the 
           lib/python/Products/ExtFile/icons/... (original ExtFile)
           lib/python/Products/CMFOptions/ExtFile/icons
           directory shold be used as icon for this file/image"""
        cat, sub = self._getMIMECatAndSub(self.content_type)
        if hasattr(self,'has_preview') and self.has_preview: cat='image'
        if self._types.has_key(cat):
            file=self._types[cat]['default']
            for item in self._types[cat].keys():
                if string.find(sub,item)>=0:
                    file=self._types[cat][item]
                    break
            return join('icons',cat,file)
        else:
            return join('icons',self._types['default'])

    ################################
    # Protected management methods #
    ################################


    #CMF-CHANGE : had to change the dtml file to have a more consistent view in the native Zope i/f
    manage_main = HTMLFile('extFileEdit', globals())       # Management Interface

    
    def manage_editExtFile(self, title='', descr='', REQUEST=None): 
        """Manage the edited values"""
        if self.title!=title: self.title=title
        if self.descr!=descr: self.descr=descr
        # update ZCatalog
        self.reindex_object()
        if REQUEST is not None:
            return MessageDialog(
                title = 'Edited',
                message = "The properties of %s have been changed!" % self.id,
                action = './manage_main',
            )
    
    manage_uploadForm = HTMLFile('extFileUpload', globals())# File upload Interface
    
    def manage_file_upload(self,  file='', content_type='', REQUEST=None):
        """Upload file from local directory"""
        if type(file) is not type(''):
            reposit=self._reposit_path()
            if UNDO_POLICY==ALWAYS_BACKUP or self.filename=='': new_fn=self._get_ufn()
            else: new_fn=self.filename
            try:
                fp=open(join(reposit, new_fn), 'wb')
                self._copy_data(file, fp)
                fp.close()
            except:
                raise 'Cannot write: ', ("%s (%s)" %(self.id, new_fn))
            self.content_type=self._get_content_type(file, file.read(100), 
                              self.id, content_type or self.content_type)
            if self.filename!='' and UNDO_POLICY==ALWAYS_BACKUP:
                old_fn=join(reposit,self.filename)
                try: os.rename(old_fn, old_fn+'.undo')
                except: pass
            self.filename=new_fn
        elif REQUEST is not None:
            return MessageDialog(title = 'Error!',
                message = "You must specify a file!", action = './manage_main',)
        if REQUEST is not None:
            return MessageDialog(title = 'Uploaded',
                message = "The file was uploaded successfully!", 
                action = './manage_main',)

    def manage_http_upload(self, url, REQUEST=None):
        """Upload file from http-server."""
        reposit=self._reposit_path()
        url=urllib.quote(url,'/:')
        createExc='Read error: '
        if UNDO_POLICY==ALWAYS_BACKUP or self.filename=='': new_fn=self._get_ufn()
        else: new_fn=self.filename
        try:
            fp_in=urllib.urlopen(url)
            fp_out=open(join(reposit, new_fn), 'wb')
            self._copy_data(fp_in, fp_out)
            fp_out.close()
        except:
            raise createExc, url
        else:
            try:
                instream=open(join(reposit, new_fn), 'rb')
                self.content_type=self._get_content_type(instream, instream.read(100),
                                  self.id, self.content_type)
                instream.close()
            except:
                if UNDO_POLICY==ALWAYS_BACKUP: os.remove(join(reposit, new_fn))
            else:
                if UNDO_POLICY==ALWAYS_BACKUP: 
                    old_fn=join(reposit,self.filename)
                    try: os.rename(old_fn, old_fn+'.undo')
                    except: pass
                self.filename=new_fn
        if REQUEST is not None:
            return MessageDialog(title = 'Uploaded',
                message = "The file was uploaded successfully!", 
                action = './manage_main',)
    
    manage_FTPget=index_html

    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP PUT requests"""
        reposit=self._reposit_path()
        self.dav__init(REQUEST, RESPONSE)
        content_type=REQUEST.get_header('content-type', None)
        instream=REQUEST['BODYFILE']
        if UNDO_POLICY==ALWAYS_BACKUP or self.filename=='': new_fn=self._get_ufn()
        else: new_fn=self.filename
        try:
            outstream=open(join(reposit,new_fn), 'wb')
            self._copy_data(instream, outstream)
            outstream.close()
        except:
            raise cantWriteExc, ("%s (%s)" %(self.id, new_fn))
        else:
            try:
                self.content_type=self._get_content_type(instream, instream.read(100),
                                  self.id, content_type or self.content_type)
            except:
                if UNDO_POLICY==ALWAYS_BACKUP: os.remove(join(reposit, new_fn))
            else:
                if UNDO_POLICY==ALWAYS_BACKUP: 
                    old_fn=join(reposit,self.filename)
                    try: os.rename(old_fn, old_fn+'.undo')
                    except: pass
                self.filename=new_fn
        RESPONSE.setStatus(204)
        return RESPONSE

    ################################
    # Private methods              #
    ################################

    def _access_permitted(self, REQUEST):
        """check if the user is allowed to download the file"""
        if (hasattr(self,'use_download_permission_check') and self.use_download_permission_check and 
            (not REQUEST.has_key('AUTHENTICATED_USER') or 
            not REQUEST['AUTHENTICATED_USER'].has_permission('Download ExtFile/ExtImage', self))): 
            return 0
        else: return 1


    def _get_content_type(self, file, body, id, content_type=None):
        """determine the mime-type"""
        headers=getattr(file, 'headers', None)
        if headers and headers.has_key('content-type'):
            content_type=headers['content-type']
        else:
            if type(body) is not type(''): body=body.data
            content_type,enc=guess_content_type(getattr(file,'filename',id),body,content_type)
        return content_type


    def _copy_data(self, instream, outstream):
        """Read binary data."""
        if type(instream) is StringType: 
            outstream.write(instream)
        else:
            try:
                instream.seek(0,2)
                size=instream.tell()        # determine the filesize in bytes
                instream.seek(0)
                blocksize=2<<16
                pos=0
                while pos<size:
                    outstream.write(instream.read(blocksize))
                    pos=pos+blocksize
                instream.seek(0)
            except:
                outstream.write(instream.read())


    def _copyfile (self, filename, outstream):
        """read binary data from file filename"""
        try:
            instream=open(filename, 'rb')
        except IOError:
            self._undo()
            try:
                instream=open(filename, 'rb')
            except IOError:
                raise IOError, ("%s (%s)" %(self.id, filename))
        try:
            self._copy_data(instream, outstream)
            instream.close()
        except IOError:
            raise IOError, ("%s (%s)" %(self.id, filename))

    def _bytetostring (self, value):
        """Convert an int-value (file-size in bytes) to an String
           with the file-size in Byte, KB or MB
        """
        bytes=float(value)
        if bytes>=1000:
            bytes=bytes/1024
            if bytes>=1000:
                bytes=bytes/1024
                typ=' MB'
            else:
                typ=' KB'
        else:
            typ=' Bytes'
        strg='%4.2f'%bytes
        strg=strg[:4]
        if strg[3]=='.': strg=strg[:3]
        strg=strg+typ
        return strg
        
    def _reposit_path(self):
        """Create path to the repository from _repository"""
        path=INSTANCE_HOME
        for item in self._repository:
            path=join(path,item)
        return path

    def _undo (self):
        """restore filename after undo or copy-paste"""
        fn=join(self._reposit_path(),self.filename)
        if not isfile(fn) and isfile(fn+'.undo'): 
            os.rename(fn+'.undo', fn)
    
    def _get_ufn(self):
        """Create a unique filename"""
        
        # CMF-CHANGE: for CMF, makes sense to have the Creator's name in the file
        #for easy cleanup of the repository. 
    
        from AccessControl import getSecurityManager
        u = str(getSecurityManager().getUser())
        udir = u[0]
        if udir not in string.lowercase[:26]:
            udir = ''

        reposit=self._reposit_path()
        time=int(DateTime().strftime('%m%d%H%M%S'))
        
        ureposit = join(reposit,udir)  #----/var/reposit/[a-z]
                
        fn=join(ureposit,"%s_%d_%s"%(u,time,self.id))
        while (isfile(fn) or isfile(fn+'.preview') 
               or isfile(fn+'.undo') or isfile(fn+'.preview.undo')):
            time=time+1
            fn=join(reposit,"%s_%d_%s"%(u,time,self.id))
        return fn
    
    ################################
    # Special management methods   #
    ################################

    def manage_afterClone(self, item, new_fn=''):
        """When a copy of the object is created (zope copy-paste-operation),
           this function is called by CopySupport.py. A copy of the external file
           is created and self.filename is changed.
        """
        reposit=self._reposit_path()
        old_fn=join(reposit,self.filename)
        if (not isfile(old_fn)) and isfile(old_fn+'.undo'): old_fn=old_fn+'.undo'
        if isfile(old_fn):
            new_fn = new_fn or self._get_ufn()
            try:
                fp=open(join(reposit,new_fn), 'wb')
                self._copyfile(old_fn, fp)
                fp.close()
            except:
                raise "Clone-Exception: ", "Couldn't clone %s" % old_fn
        self.filename = new_fn or self._get_ufn()   # change filename
        return ExtFile.inheritedAttribute ("manage_afterClone") (self, item)
        
    def manage_afterAdd(self, item, container):
        """This method is called, whenever a new object instance is created. 
           if the object is *really* new (normal case) this method does nothing.
           if the object is a result of a zope cut-paste- or rename-operation 
           (in which case the object is *not really* new) this function calls
           _undo, which restores the external file(s), whichwere renamed
           by manage_beforeDelete before.
        """
        fn=join(self._reposit_path(),self.filename)
        try:                                            # test, if file exists
            f=open(fn, 'rb')
            f.close()
        except IOError:                                 # if not, we just had a 
            self._undo()                                # rename or cut-paste!
        return ExtFile.inheritedAttribute ("manage_afterAdd") (self, item, container)

    def manage_beforeDelete(self, item, container):
        """This method is called, when the object is deleted. To support 
           undo-functionality and because this happens too, when the object 
           is moved (cut-paste) or renamed, the external file is not deleted. 
           It are just renamed to filename.undo and remains in the 
           repository, until it is deleted manually.
        """
        fn=join(self._reposit_path(),self.filename)
        try:
            if UNDO_POLICY==NO_BACKUPS: os.remove(fn)
            else: os.rename(fn, fn+'.undo')
        except OSError:
            pass
        return ExtFile.inheritedAttribute ("manage_beforeDelete") (self, item, container)

Globals.default__class_init__(ExtFile)
